//Options for these are under the MPLAB IDE Help menu/Help Contents/search for the PIC model
//internal osc with RA6&7 availabled for IO, osc not multiplied by 4
#pragma config FOSC = INTIO67, PLLCFG = OFF
//disable a bunch of features I don't use
#pragma config PRICLKEN = ON, FCMEN = OFF, IESO = OFF, PWRTEN = OFF, BOREN = OFF, WDTEN = OFF, WDTPS = 1, CCP2MX = PORTC1
//PORTB<5:0> pins are configured as digital I/O on Reset 
#pragma config PBADEN = OFF
//Config bits for stuff I don't use. Also note, RE3 can only be an input.
#pragma config CCP3MX = PORTC6, HFOFST = OFF, T3CMX = PORTC0, P2BMX = PORTC0, MCLRE = INTMCLR, STVREN = OFF, LVP = ON, XINST = ON
//debug mode off and code protection off
#pragma config DEBUG = OFF, CP0 = OFF, CP1 = OFF, CP2 = OFF, CP3 = OFF, CPB = OFF, CPD = OFF, WRT0 = OFF, WRT1 = OFF, WRT2 = OFF, WRT3 = OFF, WRTC = OFF, WRTB = OFF, WRTD = OFF, EBTR0 = OFF, EBTR1 = OFF, EBTR2 = OFF, EBTR3 = OFF, EBTRB = OFF


#include <p18f26k22.h>
#include <usart.h>
#include <adc.h>//include library ADC functions
#include <delays.h>//delay functions
#include <usart.h>//include serial functions
#include <stdlib.h>//include string conversion functions
#include <math.h>//include math functions




#define LEDPIN LATBbits.LATB5

//USART 1 will talk to servos. Tx pin is RC6, Rx pin is RC7
#define SERVOTXDDR TRISCbits.RC6//I'll control this in the program to change bus direction
#define SERVORXDDR TRISCbits.RC7//this is just for convenience; will always be an input.
#define SERVORXLAT LATCbits.LATC7
//USART 2 will talk to PC
#define TX2DDR TRISBbits.RB6
#define RX2DDR TRISBbits.RB7

#define BUFFERSIZE 32 //must be a power of 2, or the mask method below won't work.
#define BUFFERMASK (BUFFERSIZE-1)

//Now declare the variables we'll use. C18 needs variables to be declared ahead of time. It gives a syntax
//error if you declare new variables in the middle of the main() function.
unsigned char data;
unsigned char TxPCBuffer[BUFFERSIZE];//rotating buffer for Tx to PC. Required because AX-12 can xmit faster.
unsigned char TxPCHead = 0;//next index to xmit
unsigned char TxPCTailPlus = 0;//one past last to xmit;  we'll write here.
unsigned char currentID = 0;
unsigned char instruction = 0;
unsigned char msgIndex = 0;//byte where we are in the current message
char msgLengthLeft = 0;//bytes remaining, based on length byte
char outputstring[20];
char decimalstring[6];

//function prototypes
void configToXmit(void);
void configToReceive(void);
void setBaudRate(char baud);
void xmitEnqueue(char thedata);
char xmitIsEmpty(void);//actually a boolean 1 or 0 return.
char xmitPop(void);

void main (void)
{



//*****begin standard config*****
OSCCON = 0b01110111;	//bits 6-4 and 1-0 set it to use the 16MHz internal clock
//*****end standard config*****

//****begin project specific config
ANSELA = 0b00000000;	//set all ADC-capable pins to NOT be analog inputs, just digital IO
ANSELB = 0b00000000;    //if you don't do this, the inputs will just mysteriously not work!
ANSELC = 0b00000000;
TRISA = 0;//all port A as outputs
TRISBbits.TRISB5 = 0;//set LED pin, RB5, to be an output.
SERVOTXDDR = 1;//datasheet says serial port will take control of DDR, so set it to 1
TX2DDR = 1;
SERVORXDDR = 1;
SERVORXLAT = 1;//high
RX2DDR = 1;



//****end project-specific config***
//USART 1 will talk to servos
Open1USART( USART_TX_INT_OFF  &
    USART_RX_INT_OFF  &
    USART_ASYNCH_MODE &
    USART_EIGHT_BIT   &
    USART_CONT_RX     &
    USART_BRGH_HIGH,
    0                );//use 0 for 1Mbaud
//Asynchronous mode, high speed: 
//FOSC / (16 * (spbrg + 1))
//so if FOSC = 16MHz, and I want 115200 bps, I need 7.68, so round to 8

//USART 2 will talk to PC.
Open2USART( USART_TX_INT_OFF  &
    USART_RX_INT_OFF  &
    USART_ASYNCH_MODE &
    USART_EIGHT_BIT   &
    USART_CONT_RX     &
    USART_BRGH_HIGH,
    8                );

configToReceive();
while(1){
    /* immediately forward bytes from PC:
     * this does no queueing; it depends on the fact that the PC transmits slower than
     * we'll transmit to the servos.
     */
    if(DataRdy2USART()){//Got a new byte from PC, so send it.
        if(TXSTA1bits.TXEN1 ==0){//if servo bus is in receive mode
            configToXmit();//switch to xmit mode
        }
        data = Read2USART();//read from PC
        Write1USART(data);//forward to servo bus. This REQUIRES USART2 to be slower than USART1. If not, add a software buffer.
        //for simple operation we could stop here. But I need the PIC to detect end-of-packet to change bus direction,
        //so I have to parse the message bytes
        if(data == 0xFF && (msgIndex == 0 || msgIndex == 1)){//detect beginning bytes
            msgIndex++;
        }else if(msgIndex == 2){//last 2 bytes were 255; we're on ID.
            currentID = data;
            msgIndex = 3;
        }else if(msgIndex == 3){//we're on length
            msgLengthLeft = data;
            msgIndex = 4;
        }else{
            if(msgIndex==4){
                instruction = data;
            }
            if(msgIndex == 5){
                if(currentID == 255 && instruction == 3){//write message to PIC
                    setBaudRate(data);//set baud to param1
                    Write2USART(1);//acknowledge that PIC got the message
                }
            }
            msgIndex++;
            msgLengthLeft--;
            if(msgLengthLeft <= 0){//this was the last char of message
                msgIndex = 0;
                //todo: figure out how to while(stilltransmitting);
                Delay10TCY();
                while(TXSTA1bits.TRMT1 == 0){
                    //wait for xmit buffer to be empty
                }
                Delay10TCYx(1);//2.5 uS
                configToReceive();
            }
        }
    }


    /******Enqueue bytes from servos; they come too fast to send immediately like with the PC*****/
    if(TXSTA1bits.TXEN1 == 0 ){//receive mode on servo bus
        while(DataRdy1USART()){//and data has come in
            //xmitEnqueue( Read1USART() );//read a byte from servo bus serial port and enqueue it
            /* I copy-pasted xmitEnqueue function, because doing the function call is
             * too slow and was causing receive buffer overrun errors at 1Mbps, and C18
             * doesn't understand the standard C inline directive
             */
            TxPCBuffer[TxPCTailPlus] = Read1USART();
            TxPCTailPlus =(TxPCTailPlus+1) & BUFFERMASK;
        }
    }


    /******* Send an enqueued byte if the queue isn't empty *******/
    if(TxPCHead != TxPCTailPlus){//Not empty. We have something to send to PC
        if(1==TXSTA2bits.TRMT){//this bit is 1 when register is empty, meaning we can xmit to PC.
            //Write2USART( xmitPop() );//send next byte to PC
            /* I copy-pasted the xmitPop() function to be safe, because we could get
             * here during a servo bus receive, so we shouldn't delay here too long.
             */
            Write2USART(TxPCBuffer[TxPCHead]);
            TxPCHead =(TxPCHead+1) & BUFFERMASK;
        }
    }
}




}

void configToXmit(){

    if(RCSTA1bits.OERR){
        xmitEnqueue(88); // an arbitrary error code. Indicates an overrun, meaning
        xmitEnqueue(88); // the servo bus xmitted faster than we could save the data.
    }
    RCSTA1bits.CREN1 = 0;//disable receiver
    Delay10TCY();
    TXSTA1bits.TXEN1 = 1;//enable transmitter circuit
    
    LEDPIN = 1;
}

void configToReceive(){
    //SERVOTXDDR = 1;//set TX pin to input, so servos can control the bus. DOESN'T WORK because USART takes priority per table 10-4 of datasheet
                    //see datasheet PDF pg 142: EUSART overrides TRIS bits.

    //SERVORXDDR = 1;//set RX pin to input, so we can receive
    TXSTA1bits.TXEN1 = 0;//disable transmitter circuit
    Delay10TCY();
    RCSTA1bits.CREN1 = 1;//enable receiver
    //RCSTA1bits.SPEN1 = 0; //return control of data direction to TRIS bits

    /*while(DataRdy1USART()){
        Read1USART();//throw away our own stored transmissions, since receive has been monitoring the bus.
    }*/
    LEDPIN = 0;
}

void setBaudRate(char baud){
    SPBRG1 = baud;
    SPBRGH1 = 0;
}

/*Push a byte onto the end of the xmit queue
 * If this changes, change the place I copy-pasted it.
 */
void xmitEnqueue(char thedata){
    TxPCBuffer[TxPCTailPlus] = thedata;
    TxPCTailPlus =(TxPCTailPlus+1) & BUFFERMASK;//this will wrap it to beginning if necessary
    if(TxPCTailPlus == TxPCHead){//if tail==head after just adding something, we've overflowed.
        //no graceful recovery, but tell me why it failed with an (arbitrary) error code of 77, 77:
        TxPCBuffer[TxPCTailPlus] = 77;
        TxPCTailPlus =(TxPCTailPlus+1) & BUFFERMASK;
        TxPCBuffer[TxPCTailPlus] = 77;
        TxPCTailPlus =(TxPCTailPlus+1) & BUFFERMASK;
    }
}
/*true if there's nothing to xmit
 */
char xmitIsEmpty(){//actually a boolean 1 or 0 return.
    return TxPCHead == TxPCTailPlus;
}
/*Pop a byte from the beginning of xmit queue
 * If this changes, change the place I copy-pasted it.
 */
char xmitPop(){
    char returnData = TxPCBuffer[TxPCHead];
    TxPCHead =(TxPCHead+1) & BUFFERMASK;
    return returnData;
}